//---------------------------------------------------------------------------
//  TVolgaQuery - inherited from TVolgaDataSet, used for quering flat files
//  and in-memory tables
//---------------------------------------------------------------------------
//  Copyright  2000, Olga Vlasova, Russia
//  http://www.volgadb.com
//  E-mail: info@volgadb.com
//---------------------------------------------------------------------------
//  Support Kylix/Lilux platform: (c) Vit (Vitaly Nevzorov), 2004, Chicago, USA
//  http://www.delphist.com
//  E-mail: vit@vingrad.ru
//---------------------------------------------------------------------------
unit VolgaQry;

interface
uses
{$IFDEF Linux}
{$ELSE}
  Windows,
{$ENDIF Linux}
  Classes, Db, SysUtils, VolgaConst, VolgaTbl, VolgaLex;

type

  TVolgaQTable = record                 //structure for storage information about tables in request
    FullName: string;                   //table's full name + path
    Origin: string;                     //name without path and extension
    Alias: string;                      //source table alias in table list
    QCursor: TVolgaCursor;              //the table's cursor
    IsMaster: Boolean;                  //token of the main table, that must be updated
    IsFile: Boolean;                    //is query dataset a real file or opened dataset in memory
  end;
  PVolgaQTable = ^TVolgaQTable;

  TVolgaQField = record                 //structure for storage information about field in the request
    Origin: string;                     //source table name
    Alias: string;                      //source table alias in table list
    Name: string;                       //field name
    Expr: string;                       //number or string as the field value
    AsName: string;                     //field name alias, or only the name in answer
    TokenType: TVolgaTokenType;         //token type
    DataType: TVolgaFieldType;          //our field type
    Size: Word;                         //declared field length for strings
  end;
  PVolgaQField = ^TVolgaQField;

  TVolgaQuery = class;

  TQueryDataset = class(TCollectionItem)
  private
    FDataset: TVolgaDataset;
    FAlias: string;
    procedure SetDataset(Value: TVolgaDataset);
    procedure SetAlias(const Value: string);
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
  published
    property DataSet: TVolgaDataset read FDataset write SetDataset;
    property Alias: string read FAlias write SetAlias;
  end;

  TQueryDatasets = class(TCollection)
  private
    FQuery: TVolgaQuery;
    function GetItem(Index: Integer): TQueryDataset;
    procedure SetItem(Index: Integer; Value: TQueryDataset);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(Query: TVolgaQuery);
    function Add: TQueryDataset;
    function IndexOfAlias(const Alias: string): Integer;
    property Items[Index: Integer]: TQueryDataset read GetItem write SetItem; default;
  end;

  //********* TVolgaQuery - request for flat files ******************************

  TVolgaQuery = class(TVolgaCustomQuery)
  private
    FParser: TVolgaParser;
    FTableList: TList;
    FFieldList: TList;
    FJoinList: TList;
    FLinkList: TList;
    FWhereList: TList;
    FRecs: TList;
    FUniqueFI: TList;
    FGroupBy: Boolean;                  //is grouping of the answer necessary?
    FDistinct: Boolean;                 //SELECT DISTINCT token
    FQueryDatasets: TQueryDatasets;
    procedure PrepareTables(LoadRecords: Boolean); //open the tables, that are in request
    procedure TransferLists;            //move lists from the parser to our lists
    procedure InitAnswer;               //prepare the answer structure
    procedure AliasWhereList(aList: TList; iAlias: string);
    function GetTable(tn: string; var i: integer): PVolgaQTable;
    procedure CheckSelectFields;        //check the presence of fields, listed in SELECT in tables
    procedure CheckJoinFields;
    procedure CheckWhereFields;
    function PrepareAnswLayout: string;
    procedure FillAnswerPtr;
    procedure FillAnswer;               //fill in the answer with data
    procedure InsertMaster(T: PVolgaQTable; pRec: PVolgaRecord);
    procedure DeleteMaster(T: PVolgaQTable; pRec: PVolgaRecord);
    procedure UpdateMaster(T: PVolgaQTable; pRec: PVolgaRecord);
    function FindMaster(T: PVolgaQTable; pRec: PVolgaRecord): PVolgaRecord;
    procedure PrepareNewBuf(T: PVolgaQTable; pRec: PVolgaRecord; NewBuf: PChar);
    procedure GroupAnswer;
    procedure SetDatasets(const Value: TQueryDatasets);
  protected
    procedure SetOrigin(const Value: string); override;
    procedure InternalLoadData; override; //method of the real data loading
    procedure InternalApplyUpdates; override; //method of the real data saving
    procedure Prepare; override;
    procedure UnPrepare; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AddDataset(Dataset: TVolgaDataset; const Alias: string);
    procedure ClearDatasets;
  published
    property QueryDatasets: TQueryDatasets read FQueryDatasets write SetDatasets;
    property Password;
  end;

implementation

//******************************************************************************
//          TVolgaQuery - request for flat files
//******************************************************************************

constructor TVolgaQuery.Create(AOwner: TComponent);
begin
  inherited;
  FGroupBy := false;
  FDistinct := false;
  FQueryDatasets := TQueryDatasets.Create(Self);
end;

procedure TVolgaQuery.Prepare;
//parse the SQL text and prepare it's execution
var AText: string;
begin
  if Prepared then Exit;
  AText := GetPreparedSQL(DecimalSeparator, DateSeparator, ShortDateFormat);
  FParser := TVolgaParser.Create;       //create parser for expression
  try
    FDistinct := FParser.PrepareSQLText(AText);
  except
    FParser.Free;
    FParser := nil;
    raise;
  end;
  FTableList := TList.Create;
  FFieldList := TList.Create;
  FWhereList := TList.Create;
  FJoinList := TList.Create;
  FLinkList := TList.Create;
  FRecs := TList.Create;
  try
    try
      TransferLists;                    //move data into the TVolgaQuery list
      FPrepared := true;                //if there were no errors before
    except
      raise;
    end;
  finally
    FParser.Free;                       //now we can free parser
    FParser := nil;
  end;
end;

procedure TVolgaQuery.UnPrepare;
//SQL is changed or request is closed
var i: integer;
  T: PVolgaQTable;
  F: PVolgaQField;
  J: PVolgaQJoin;
  C: PVolgaCField;
begin
  if not Prepared then Exit;
  FPrepared := false;
  FGroupBy := false;
  FDistinct := false;
  if FRecs <> nil then
  begin
    for i := 0 to FRecs.Count - 1 do
    begin
      if FRecs.Items[i] <> nil then FreeMem(FRecs.Items[i]);
      FRecs.Items[i] := nil;
    end;
    FRecs.Free;
    FRecs := nil;
  end;
  if FTableList <> nil then
  begin
    for i := 0 to FTableList.Count - 1 do
    begin
      T := PVolgaQTable(FTableList.Items[i]);
      if T <> nil then
      begin
        if T^.IsFile and (T^.QCursor <> nil) then
          T^.QCursor.Free;
        T^.QCursor := nil;
        Dispose(T);
      end;
    end;
    FTableList.Free;
    FTableList := nil;
  end;
  if FFieldList <> nil then
  begin
    for i := 0 to FFieldList.Count - 1 do
    begin
      F := PVolgaQField(FFieldList.Items[i]);
      if F <> nil then Dispose(F);
    end;
    FFieldList.Free;
    FFieldList := nil;
  end;
  if FJoinList <> nil then
  begin
    for i := 0 to FJoinList.Count - 1 do
    begin
      J := PVolgaQJoin(FJoinList.Items[i]);
      if J <> nil then Dispose(J);
    end;
    FJoinList.Free;
    FJoinList := nil;
  end;
  if FLinkList <> nil then
  begin
    for i := 0 to FLinkList.Count - 1 do
    begin
      J := PVolgaQJoin(FLinkList.Items[i]);
      if J <> nil then Dispose(J);
    end;
    FLinkList.Free;
    FLinkList := nil;
  end;
  if FWhereList <> nil then
  begin
    for i := 0 to FWhereList.Count - 1 do
    begin
      C := PVolgaCField(FWhereList.Items[i]);
      if C <> nil then
      begin
        if C^.Maska <> nil then C^.Maska.Free;
        Dispose(C);
      end;
    end;
    FWhereList.Free;
    FWhereList := nil;
  end;
end;

procedure TVolgaQuery.TransferLists;
//move data into the TVolgaQuery list
var FS: PVolgaQSource;
  FT: PVolgaQTable;
  FF: PVolgaQField;
  FC: PVolgaQColumn;
  FJ: PVolgaQJoin;
  FW, FW1: PVolgaCField;
  i, j, p: integer;
  fp: string;
begin
  for i := 0 to FParser.SourceList.Count - 1 do
  begin
    FS := PVolgaQSource(FParser.SourceList.Items[i]);
    New(FT);
    if IsValidIdent(FS^.Alias) then
      FT^.Alias := FS^.Alias
    else
      FT^.Alias := 'XXX';
    p := FQueryDatasets.IndexOfAlias(FS^.Alias);
    if p >= 0 then
    begin
      FT^.IsFile := false;
      FT^.Origin := '';                 //FQueryDatasets[p].Dataset.Origin;
      FT^.FullName := '';
      FT^.QCursor := GetCursor(FQueryDatasets[p].Dataset);
    end
    else
    begin
      FT^.IsFile := true;
      if Database <> nil then
        fp := Database.DatabasePath
      else
        fp := '';
      if fp > '' then
        FT^.FullName := fp + FS^.Origin
      else
        FT^.FullName := FS^.Origin;
      if not FileExists(FT^.FullName) then
        FT^.FullName := FT^.FullName + '.d';
      FT^.Origin := FS^.Origin;
      FT^.QCursor := nil;
    end;
    //future table for updating  can be only the first!?
    FT^.IsMaster := (AnsiCompareText(FT^.Alias, Origin) = 0) and (UniqueFieldNames >
      '') and FT^.IsFile;
    FTableList.Add(FT);
  end;
  //move master-table to the first place
  if FTableList.Count > 1 then
    for i := 0 to FTableList.Count - 1 do
      if PVolgaQTable(FTableList.Items[i])^.IsMaster then
      begin
        FTableList.Move(i, 0);          //move
        Break;
      end;
  for i := 0 to FParser.ColumnList.Count - 1 do
  begin
    FC := PVolgaQColumn(FParser.ColumnList.Items[i]);
    for j := 0 to FFieldList.Count - 1 do
      if FC^.AsName = PVolgaQField(FFieldList.Items[j])^.AsName then
        FC^.AsName := FC^.AsName + '_1';
    New(FF);
    FF^.Name := FC^.Name;
    FF^.Alias := FC^.Alias;
    FF^.AsName := FC^.AsName;
    FF^.DataType := vftUnknown;
    FF^.TokenType := FC^.TokenType;
    if FDistinct then
      if FC^.TokenType = vttAster then
        FF^.TokenType := vttAsterGroup
      else
        FF^.TokenType := vttGroup;
    if FF^.TokenType in [vttCount, vttSum, vttMax, vttMin, vttAvg, vttGroup, vttAsterGroup] then
      FGroupBy := true;                 //"request with groupinng" token
    FF^.Size := 0;
    if (FF^.Name = FF^.AsName) and (FF^.TokenType in [vttCount, vttSum, vttMax,
      vttMin, vttAvg]) then
      case FF^.TokenType of
        vttCount: FF^.AsName := 'Count_' + FF^.Name;
        vttSum: FF^.AsName := 'Sum_' + FF^.Name;
        vttMax: FF^.AsName := 'Max_' + FF^.Name;
        vttMin: FF^.AsName := 'Min_' + FF^.Name;
        vttAvg: FF^.AsName := 'Avg_' + FF^.Name;
      end;
    FFieldList.Add(FF);
  end;
  if FGroupBy then                      //check whether all fields are listed either in GROUP BY or in functions
    for i := 0 to FFieldList.Count - 1 do
    begin
      FF := PVolgaQField(FFieldList.Items[i]);
      if not (FF^.TokenType in [vttCount, vttSum, vttMax, vttMin, vttAvg, vttGroup, vttAsterGroup]) then
        DatabaseErrorFmt(V_NOTGROUPFIELD, [FF^.Alias + '.' + FF^.AsName]);
    end;
  for i := 0 to FParser.WhereList.Count - 1 do
  begin
    FW1 := PVolgaCField(FParser.WhereList.Items[i]);
    if (FW1^.Alias2 = '') or (AnsiCompareText(FW1^.Alias, FW1^.Alias2) = 0) then
    begin
      New(FW);                          //move only comparison conditions, and not link conditions!!!
      FW^.Name := FW1^.Name;
      FW^.Alias := FW1^.Alias;
      FW^.Name2 := FW1^.Name2;
      FW^.Alias2 := FW1^.Alias2;
      FW^.Clause := FW1^.Clause;
      FW^.CompText := FW1^.CompText;
      FW^.TokenType := FW1^.TokenType;  //TokenType of the first operand
      FW^.TokenType2 := FW1^.TokenType2; //TokenType of the second operand
      FW^.Maska := nil;
      FWhereList.Add(FW);
    end
    else if (FW1^.Clause = vcEqual) or (FW1^.Clause = vcLeftEqual) then
    begin                               //it is link condition
      New(FJ);
      FJ^.Name := FW1^.Name;
      FJ^.Alias := FW1^.Alias;
      FJ^.Name2 := FW1^.Name2;
      FJ^.Alias2 := FW1^.Alias2;
      if FW1^.Clause = vcEqual then
        FJ^.JoinType := vjInner
      else
        FJ^.JoinType := vjLeft;
      FJoinList.Add(FJ);                //check the presence of such field in table!!!!!!!!
    end
    else
      DatabaseError(V_UNSUPPORTJOIN, Self);
  end;
end;

procedure TVolgaQuery.InternalLoadData;
//prepare InBuf and InLen, then InternalLoad is called
begin
  Prepare;
  PrepareTables(false);                 //prepare information about the opened tables structure
  CheckSelectFields;                    //make sure that there are fields from the request in the real table and enter their type
  CheckWhereFields;                     //make sure that there are fields from the conditions in the real table
  CheckJoinFields;                      //make sure that there are fields from the links in the real table
  PrepareTables(true);                  //here we open tables, participating in request
  InitAnswer;                           //here we create the empty cursor of the answer and call InternalLoad
  FillAnswerPtr;                        //here we fill in the pointer  on the answer
  FillAnswer;                           //here we fill in the answer
end;

procedure TVolgaQuery.PrepareTables(LoadRecords: Boolean);
//load all tables from the request in accordance with filters
var f1: TFileStream;
  m1: TMemoryStream;
  T: PVolgaQTable;
  i: integer;
  msg: string;
  Success: Boolean;
  FiltList: TList;
begin
  for i := 0 to FTableList.Count - 1 do //load ALL tables from the request, perhaps partly
  begin
    T := PVolgaQTable(FTableList.Items[i]);
    if FGroupBy or FDistinct then
      T^.IsMaster := false;             //no updates when grouping !!!!!!!
    if T^.IsFile then
    begin
      if not FileExists(T^.FullName) then
        DatabaseErrorFmt(V_OPENFILE, [T^.FullName], Self);
      if T^.QCursor = nil then
        T^.QCursor := TVolgaCursor.Create
      else
        T^.QCursor.Clear(true);
      if Password > '' then T^.QCursor.Password := Password;
      if LoadRecords then
      begin
      //assign the filter for cursor but not for table, that must be edited!
        FiltList := TList.Create;
        AliasWhereList(FiltList, T^.Alias);
        AcceptFilterList := FiltList;
        if (FiltList.Count > 0) and not T^.IsMaster then
          T^.QCursor.FilterClause := AcceptFilter;
        m1 := TMemoryStream.Create;
        m1.LoadFromFile(T^.FullName);
        Success := T^.QCursor.LoadFromStream(m1, Msg); //load the whole file from the stream
        m1.Free;
        FiltList.Clear;
        FiltList.Free;
        T^.QCursor.FilterClause := nil;
      end
      else
      begin
        f1 := TFileStream.Create(T^.FullName, fmOpenRead); //here we don't read the whole file, but only header
        Success := T^.QCursor.LoadFieldInfo(f1, Msg); //load fields info
        f1.Free;
      end;
      if not Success then
      begin
        T^.QCursor.Free;
        T^.QCursor := nil;
        DatabaseError(V_OPENTABLE + Msg, Self);
      end;
    end;
  end;
end;

function TVolgaQuery.GetTable(tn: string; var i: integer): PVolgaQTable;
var j: integer;
begin                                   //consider that table will be found in any case
  Result := nil;
  for j := 0 to FTableList.Count - 1 do
  begin
    i := j;
    Result := PVolgaQTable(FTableList.Items[j]);
    if AnsiCompareText(Result^.Alias, tn) = 0 then Break;
  end;
end;

procedure TVolgaQuery.CheckWhereFields;
//make sure that there are fields from condition in the real table and enter the parameters
var i, k: integer;
  F: PVolgaCField;
  QC: PVolgaQTable;
begin
  if FWhereList.Count = 0 then Exit;
  for i := 0 to FWhereList.Count - 1 do
  begin
    F := PVolgaCField(FWhereList.Items[i]); //we take the field
    QC := GetTable(F^.Alias, k);
    case F^.TokenType of
      vttIdent, vttQuoted:
        if QC^.QCursor.FindFieldInfo(F^.Name) = nil then
          DatabaseErrorFmt(V_FIELDNOTFOUND, [F^.Name, F^.Alias],
            Self);
    end;
    case F^.TokenType2 of
      vttIdent, vttQuoted:
        begin
          QC := GetTable(F^.Alias2, k);
          if (QC^.QCursor.FindFieldInfo(F^.Name2) = nil) or (F^.Alias2 = '') then
          begin
            if F^.TokenType2 = vttIdent then
              DatabaseErrorFmt(V_FIELDNOTFOUND, [F^.Name2, F^.Alias2], Self);
            F^.CompText := F^.Name2;
            if F^.Clause = vcLike then
              F^.Maska := TVolgaMask.Create(F^.CompText);
          end;
        end;
      vttNumber, vttNull, vttTrue, vttFalse:
        F^.CompText := F^.Name2;
    end;
  end;
end;

procedure TVolgaQuery.CheckSelectFields;
//make sure that there are fields from the request in the real table and enter their type
var i, k, l: integer;
  F, FF: PVolgaQField;
  QC: PVolgaQTable;
  FI: PVolgaField;
  IsFound: Boolean;
  tt: TVolgaTokenType;
begin
  for i := 0 to FFieldList.Count - 1 do
  begin
    F := PVolgaQField(FFieldList.Items[i]); //we take the field
    tt := F^.TokenType;
    case tt of
      vttIdent, vttQuoted, vttCount, vttSum, vttMax, vttMin, vttAvg, vttGroup:
        if F^.DataType = vftUnknown then
        begin
          IsFound := false;             //look for it in all opened cursors
          QC := GetTable(F^.Alias, l);
          FI := QC^.QCursor.FindFieldInfo(F^.Name);
          if FI <> nil then
          begin
            IsFound := true;
            F^.Origin := QC^.Origin;
            if F^.TokenType <> vttCount then
            begin
              F^.DataType := FI^.DataType;
              F^.Size := FI^.Size;
            end
            else
            begin
              F^.DataType := vftInteger;
              F^.Size := 0;
            end;
          end;
          if not IsFound then
            if F^.TokenType = vttIdent then
              DatabaseErrorFmt(V_FIELDNOTFOUND, [F^.Name, F^.Alias], Self)
            else                        //it is NOT the name of the field in quote, but simple string!
            begin
              F^.Origin := 'XXX';
              F^.Alias := '';
              F^.Expr := F^.Name;
              if Length(F^.Name) <= MAXSLEN then
                F^.DataType := vftSmallStr
              else
                F^.DataType := vftString;
              F^.Size := Length(F^.Name);
            end;
        end;
      vttNumber:
        begin
          F^.Origin := 'XXX';
          F^.Alias := '';
          F^.Expr := F^.Name;
          F^.DataType := vftDouble;     //always float??????????????
          F^.Size := 0;
        end;
      vttAster, vttAsterGroup:
        begin                           //if the field - an asterisk (all field of the table)
          QC := GetTable(F^.Alias, l);
          Dispose(F);
          FFieldList.Items[i] := nil;   //deleting asterisk
          //and listing all fields instead of asterisk
          for k := 0 to QC^.QCursor.FieldCount - 1 do
          begin
            New(FF);
            FF^.Origin := QC^.Origin;
            FF^.Alias := QC^.Alias;
            FF^.Name := QC^.QCursor.FieldInfo[k]^.Name;
            FF^.AsName := FF^.Name;
            FF^.DataType := QC^.QCursor.FieldInfo[k]^.DataType;
            FF^.Size := QC^.QCursor.FieldInfo[k]^.Size;
            if tt = vttAsterGroup then
              FF^.TokenType := vttGroup
            else
              FF^.TokenType := vttIdent;
            FFieldList.Add(FF);
          end;
        end;                            //continue browsing the tables
//    else //error - invalid token
    end;
  end;
  FFieldList.Pack;
end;

procedure TVolgaQuery.CheckJoinFields;
//make sure that Join fields belong to the specified tables?
var i, j, l: integer;
  F, pLink: PVolgaQJoin;
  QC: PVolgaQTable;
  T: PVolgaQTable;
  secAlias: string;
label Lab1;
begin
  if FJoinList.Count = 0 then Exit;
  for i := 0 to FJoinList.Count - 1 do
  begin
    F := PVolgaQJoin(FJoinList.Items[i]); //we take the field
    QC := GetTable(F^.Alias, l);
    if QC^.QCursor.FindFieldInfo(F^.Name) = nil then
      DatabaseErrorFmt(V_FIELDNOTFOUND, [F^.Name, F^.Alias], Self);
    QC := GetTable(F^.Alias2, l);
    if QC^.QCursor.FindFieldInfo(F^.Name2) = nil then
      DatabaseErrorFmt(V_FIELDNOTFOUND, [F^.Name2, F^.Alias2], Self);
  end;
  //we create the links list
  for i := 0 to FTableList.Count - 1 do
  begin
    T := PVolgaQTable(FTableList.Items[i]);
    Lab1:                               //looking for other links of this table with the other tables
    secAlias := '';
    New(pLink);
    pLink^.Alias := T^.Alias;
    pLink^.Alias2 := '';
    pLink^.Name := '';
    pLink^.Name2 := '';
    pLink^.JoinType := vjInner;
    for j := 0 to FJoinList.Count - 1 do
    begin
      F := PVolgaQJoin(FJoinList.Items[j]); //we take the field
      if AnsiCompareText(F^.Alias, pLink^.Alias) = 0 then
        if (secAlias = '') or (AnsiCompareText(secAlias, F^.Alias2) = 0) then
        begin
          pLink^.Alias2 := F^.Alias2;
          pLink^.Name := pLink^.Name + F^.Name + ';';
          pLink^.Name2 := pLink^.Name2 + F^.Name2 + ';';
          if F^.JoinType = vjLeft then  //if it we have met at list one left join in links, than the whole Join is Left
            pLink^.JoinType := vjLeft;
          secAlias := pLink^.Alias2;
          F^.Alias := '';
          F^.Alias2 := '';
        end
        else
      else if F^.Alias2 = pLink^.Alias then
        if (secAlias = '') or (AnsiCompareText(secAlias, F^.Alias) = 0) then
        begin
          pLink^.Alias2 := F^.Alias;
          pLink^.Name := pLink^.Name + F^.Name2 + ';';
          pLink^.Name2 := pLink^.Name2 + F^.Name + ';';
          if F^.JoinType = vjLeft then  //if it we have met at list one left join in links, than the whole Join is Left
            pLink^.JoinType := vjLeft;
          F^.Alias := '';
          F^.Alias2 := '';
          secAlias := pLink^.Alias2;
        end
        else
          ;
    end;
    if pLink^.Name > '' then
      FLinkList.Add(pLink)
    else
      Dispose(pLink);
    if secAlias > '' then
      goto Lab1;                        //looking for other links of this table with the other tables
  end;
end;

procedure TVolgaQuery.InitAnswer;
//prepare the structure of the answer and create the empty cursor
var s: string;
  rec, len: integer;
  typ: char;
  Stream: TMemoryStream;
begin
  s := PrepareAnswLayout;               //preparing string of the fields description

  Stream := TMemoryStream.Create;
  len := SIZEINT * 4 + 1 + Length(s);
  Stream.SetSize(len);
  Stream.Write(len, SIZEINT);           //buffer length
  typ := 'S';
  Stream.Write(typ, 1);                 //success
  rec := 0;
  Stream.Write(rec, SIZEINT);           //record count:= 0
  rec := FFieldList.Count;              //selecting  records if their AsName property >'' !!!!!!!!!!!!!!!!!!!
  Stream.Write(rec, SIZEINT);           //fields count
  rec := Length(s);
  Stream.Write(rec, SIZEINT);           //header length
  Stream.Write(PChar(s)^, Length(s));   //information about the fields
  Stream.Position := 0;                 //from the very begining of the stream
  InternalLoad(Stream);                 //load empty structure of the answer and crete the cursor
end;

function TVolgaQuery.PrepareAnswLayout: string;
//create fields descriptions of the answer
var
  F: PVolgaQField;
  dlina: string;
  typ: char;
  i: integer;
begin
  Result := '';
  for i := 0 to FFieldList.Count - 1 do
  begin
    F := PVolgaQField(FFieldList.Items[i]);
    if F^.AsName > '' then
    begin
      dlina := '';
      if F^.DataType = vftAutoInc then  //for autoinc fields result field is an integer
        F^.DataType := vftInteger;
      typ := GetCharType(F^.DataType, F^.Size);
      if F^.DataType = vftString then dlina := IntToStr(F^.Size);
      if IsValidIdent(F^.Origin) then
        Result := Result + F^.Origin
      else
        Result := Result + F^.Alias;
      Result := Result + '.' + F^.AsName + ' ' + typ + dlina + FieldSeparator;
    end;
  end;
  Result := Result + RecordSeparator;
end;

procedure TVolgaQuery.AliasWhereList(aList: TList; iAlias: string);
//create the list of the fields from the one table for filter
var i: integer;
  FW: PVolgaCField;
begin
  aList.Clear;
  for i := 0 to FWhereList.Count - 1 do
  begin
    FW := PVolgaCField(FWhereList.Items[i]);
    if AnsiCompareText(FW^.Alias, iAlias) = 0 then aList.Add(FW);
  end;
end;

var KeyFI1: TList;
  ali1: integer;                        //variables, required for sorting

function SQLSort(Item1, Item2: Pointer): integer;
var cbuf1, cbuf2: PChar;
begin
  Result := 0;
  Move((PChar(Item1) + SIZECHAR * ali1)^, cbuf1, SIZECHAR); //read the address
  Move((PChar(Item2) + SIZECHAR * ali1)^, cbuf2, SIZECHAR); //read the address
  if (cbuf1 = nil) and (cbuf2 = nil) then Exit;
  if cbuf1 = nil then
  begin
    Result := -1;
    Exit;
  end
  else if cbuf2 = nil then
  begin
    Result := 1;
    Exit;
  end;
  Result := CompareRec(cbuf1, cbuf2, KeyFI1);
end;

procedure TVolgaQuery.FillAnswerPtr;
//Fill in the answer data
var pRec: PVolgaRecord;
  IsFirst, IsFound, Accept: Boolean;
  T1, T2: PVolgaQTable;
  L: PVolgaQJoin;
  FI1, FI2: PVolgaField;
  {KeyFI1,} KeyFI2, FiltList: TList;
  iRow, iLink, jKey, start, ali2, len, compar, keystart: integer;
  CurBuf, compbuf1, compbuf2, NewBuf: PChar;
label Lab1;
begin
  if FLinkList.Count <> FTableList.Count - 1 then
    DatabaseError(V_ERRLINKCOUNT);

  //offset in the pointers array = (index FTableList) * SIZECHAR
  len := SIZECHAR * FTableList.Count;
  T1 := PVolgaQTable(FTableList.Items[0]);
  FRecs.Capacity := T1^.QCursor.RecordCount;
  FiltList := TList.Create;
  if T1^.IsMaster or not T1^.IsFile then
  begin
    AliasWhereList(FiltList, T1^.Alias);
    AcceptFilterList := FiltList;       //list of the fields for comparison
  end;
  for iRow := 0 to T1^.QCursor.RecordCount - 1 do
  begin
    pRec := T1^.QCursor.RecData[iRow];
    if (pRec^.Status = vusAbsent) or (pRec^.Status = vusDeleted) then
      Continue;
    Accept := true;
    if T1^.IsMaster or not T1^.IsFile then //we haven't applied filter for master-table yet, so let's apply
      AcceptFilter(T1^.QCursor, pRec, Accept);
    if not Accept then
      Continue;
    NewBuf := AllocMem(len);            //the pointers count will be the same as tables count
    compbuf1 := pRec^.Data;
    Move(compbuf1, NewBuf^, SIZECHAR);  //moving address to the buffer (the offset is 0 for the 1-st table)
    FRecs.Add(NewBuf);
  end;
  if T1^.IsMaster or not T1^.IsFile then
  begin
    FiltList.Clear;
    T1^.QCursor.FilterClause := nil;
  end;
  FiltList.Free;
  if FTableList.Count = 1 then Exit;    //we have finished the work with single table in request

  //continue with links
  for iLink := 0 to FLinkList.Count - 1 do
  begin
    L := PVolgaQJoin(FLinkList.Items[iLink]); //first link
    T1 := GetTable(L.Alias, ali1);      //first table of the link
    T2 := GetTable(L.Alias2, ali2);     //second table of the link
    KeyFI1 := TList.Create;
    KeyFI2 := TList.Create;
    GetFldInfoList(T1^.QCursor, KeyFI1, L^.Name); //fields list in source table
    GetFldInfoList(T2^.QCursor, KeyFI2, L^.Name2); //fields list in linking table

    //sort both tables by JOIN's key fields
    if FRecs.Count > 0 then FRecs.Sort(SQLSort);
    T2^.QCursor.Sort(L^.Name2, false);

    //for all records of the 1-st and 2-nd tables
    iRow := 0;                          //record number for 1-st table
    start := 0;                         //record number for 2-nd table
    keystart := 0;                      //record number, where the next value of the key in the 2-nd table has changed
    while iRow < FRecs.Count do
    begin
      CurBuf := PChar(FRecs.Items[iRow]); //1-st table current string buffer
      Move((CurBuf + SIZECHAR * ali1)^, compbuf1, SIZECHAR); //read the address of the record from the 1-st table
      //may be we will not find it, if it was left join
      if compbuf1 = nil then
      begin
        Inc(iRow);                      //go to next record
        Continue;
      end;
      IsFirst := true;
      Lab1:                             //repeat searching for the other values
      IsFound := false;
      pRec := nil;
      while not IsFound and (start < T2^.QCursor.RecordCount) do
      begin                             //with the FOLLOWING records of the 2-nd table till equality Un-is not necessary!
        pRec := T2^.QCursor.RecData[start];
        if (pRec^.Status = vusAbsent) or (pRec^.Status = vusDeleted) then
        begin
          Inc(start);                    //go to searching in the next record
          Continue;
        end;
        jKey := 0;
        compar := 0;
        while (compar = 0) and (jKey < KeyFI2.Count) do
        begin
          FI1 := PVolgaField(KeyFI1.Items[jKey]); //field, that we use for searching in the 1-st table
          FI2 := PVolgaField(KeyFI2.Items[jKey]); //field, that we use for searching in the 2-nd table
          compar := compar + CompareBuf(compbuf1 + FI1^.Offset,
            pRec^.Data + FI2^.Offset, FI1, false);
          Inc(jKey);                    //fields count in the one link
        end;
        IsFound := compar = 0;
        if compar > 0 then
          Inc(start)                    //go to searching in the next record
        else
          break;                        //equality  or the current key is smaller then sought
      end;
      //in the START variable - founded record's index
      if IsFound then
      begin
        compbuf2 := pRec^.Data; //the found record's address
        if IsFirst then
        begin
          keystart := start;            //the key has changed
          Move(compbuf2, (CurBuf + ali2 * SIZECHAR)^, SIZECHAR); //the found string in the second table
          IsFirst := false;
        end
        else
        begin                           //it is necessary to duplicate the string!!!
          NewBuf := AllocMem(len);
          Move(CurBuf^, NewBuf^, SIZECHAR * FTableList.Count); //copy the pointers of the previous record
          Move(compbuf2, (NewBuf + ali2 * SIZECHAR)^, SIZECHAR); //the found string in the second table
          Inc(iRow);                    //insert after the current????
          FRecs.Insert(iRow, NewBuf);
        end;
        Inc(start);
        goto Lab1;                      //continue checking (one to many) from the record number START
      end
      else if IsFirst and (L^.JoinType = vjInner) then //record was not found at all!!
      begin
        Dispose(CurBuf);                //clearing is not necessary for LEFT JOINs
        if iRow = FRecs.Count - 1 then  //is  it the last record?
        begin
          FRecs.Delete(iRow);           //reduce the number of the records in the result
          break;
        end
        else
        begin
          FRecs.Delete(iRow);           //reduce the number of the records in the result
          Continue;
        end;
      end;
      if not IsFirst then start := keystart; //repeat
      Inc(iRow);                        //go to the following record
    end;                                //end of the records
    //when we have finished the work with this link
    KeyFI1.Free;
    KeyFI2.Free;
  end;
end;

procedure TVolgaQuery.FillAnswer;
//fill the cursor of the answer with the prepared pointer
var Ofs: array of array of integer;
  T: PVolgaQTable;
  FI: PVolgaField;
  iTab, iRow, iCol, j, ali: integer;
  compBuf, sourBuf, NewBuf: PChar;
  aName, Msg: string;
  F: PVolgaQField;
begin
  SetLength(Ofs, FTableList.Count, FCursor.FieldCount);
  for iTab := 0 to FTableList.Count - 1 do
    for iCol := 0 to FCursor.FieldCount - 1 do
      Ofs[iTab, iCol] := -1;
  //fill the result fields offset with ( the fields of the requested table
  for iCol := 0 to FCursor.FieldCount - 1 do
  begin
    aName := PVolgaField(FCursor.FieldInfo[iCol])^.Name; //field name in answer
    for j := 0 to FFieldList.Count - 1 do
    begin
      F := PVolgaQField(FFieldList.Items[j]);
      if (AnsiCompareText(F^.AsName, aName) = 0) and (F^.TokenType <> vttCount) then
      begin
        T := GetTable(F^.Alias, ali);   //source table
        FI := T^.QCursor.FindFieldInfo(F^.Name);
        if FI <> nil then Ofs[ali, iCol] := FI^.Offset; //offset of the field in source table
        //else literal!!!!!!!!!!!!!
        break;
      end;
    end;
  end;
  for iRow := 0 to FRecs.Count - 1 do
  begin
    compbuf := PChar(FRecs.Items[iRow]);
    NewBuf := AllocMem(FCursor.RecSize);
    for iTab := 0 to FTableList.Count - 1 do
      for iCol := 0 to FCursor.FieldCount - 1 do
      begin
        if Ofs[iTab, iCol] >= 0 then
        begin                           //-1 is the literal as value of the field!!!!!!!!!!!!
          FI := FCursor.FieldInfo[iCol];
          Move((compbuf + SIZECHAR * iTab)^, sourBuf, SIZECHAR); //read the address
          if sourbuf <> nil then
          begin
            sourBuf := sourBuf + Ofs[iTab, iCol]; //add the offset to the address
            case FI^.DataType of
              vftString:
                CopyLongString(sourBuf, NewBuf + FI^.Offset);
              vftMemo, vftFmtMemo, vftGraphic, vftBlob:
                CopyBlobField(sourBuf, NewBuf + FI^.Offset);
            else
              Move(sourBuf^, (NewBuf + FI^.Offset)^, FI^.DataSize);
            end;
          end;
        end;
      end;
    try
      if not FCursor.SaveRecord(NewBuf, nil, Msg) then
        DatabaseError(Msg, Self);
    finally
      FCursor.ClearData(NewBuf);
      //FreeMem(NewBuf);
    end;
  end;
  SetLength(Ofs, 0, 0);                 //clear the array
  //close the temporary cursors after filling
  for iTab := 0 to FTableList.Count - 1 do
  begin
    T := PVolgaQTable(FTableList.Items[iTab]);
    if T^.IsMaster then Continue;       //do not close master-table!!!
    if T^.IsFile and (T^.QCursor <> nil) then
      T^.QCursor.Free;
    T^.QCursor := nil;
  end;
  if FGroupBy or FDistinct then
    GroupAnswer;                        //here we apply the grouping and static functions!!!!
  FCursor.CommitUpdates;                //transform all records from Inserted into the Unmodified
end;

procedure TVolgaQuery.GroupAnswer;
//grouping and static functions, DISTINCT
var GroupFlds: string;
  i, j, avg: integer;
  pRec1, pRec2: PVolgaRecord;
  GroupL, AggL: TList;
  F: PVolgaQField;
  FI: PVolgaField;
  HasAvg: Boolean;
  ri1, ri2: integer;
  rd1, rd2: double;

  procedure SumBuf;
  var key1, key2: PChar;
  begin
    key1 := pRec1^.Data + FI^.Offset;
    key2 := pRec2^.Data + FI^.Offset;
    case FI^.DataType of                //field type dependent comparison
      vftInteger:
        begin
          Move(key1^, ri1, SIZEINT);
          Move(key2^, ri2, SIZEINT);
          ri1 := ri1 + ri2;
          Move(ri1, key2^, SIZEINT);
        end;
      vftDouble, vftCurrency:
        begin
          Move(key1^, rd1, SIZEDOUB);
          Move(key2^, rd2, SIZEDOUB);
          rd1 := rd1 + rd2;
          Move(rd1, key2^, SIZEDOUB);
        end;
    end;
  end;

  procedure SaveAvgCount;
  var j: integer;
  begin
    for j := 0 to AggL.Count - 1 do
    begin
      F := PVolgaQField(AggL.Items[j]);
      FI := FCursor.FindFieldInfo(F^.AsName);
      if FI <> nil then
        if (F^.TokenType = vttAvg) then
        begin
          Move((pRec1^.Data + FI^.Offset)^, rd1, SIZEDOUB);
          rd1 := rd1 / avg;
          Move(rd1, (pRec1^.Data + FI^.Offset)^, SIZEDOUB);
        end
        else if (F^.TokenType = vttCount) then
        begin
          Move(avg, (pRec1^.Data + FI^.Offset)^, SIZEINT);
        end;
    end;
  end;

begin
  GroupFlds := '';                      //fields list for grouping
  AggL := TList.Create;
  HasAvg := false;                      //"there is static function AVG/COUNT" token
  for i := 0 to FFieldList.Count - 1 do
  begin
    F := PVolgaQField(FFieldList.Items[i]);
    HasAvg := HasAvg or (F^.TokenType = vttAvg) or (F^.TokenType = vttCount);
    if F^.TokenType = vttGroup then
      GroupFlds := GroupFlds + F^.AsName + ';'
    else
      AggL.Add(F);
  end;
  FCursor.Sort(GroupFlds, false);
  GroupL := TList.Create;
  GetFldInfoList(FCursor, GroupL, GroupFlds); //fields list for grouping
  avg := 1;
  for i := 0 to FCursor.RecordCount - 2 do
  begin
    pRec1 := FCursor.RecData[i];
    pRec2 := FCursor.RecData[i + 1];
    //here we compare two nearby records:
    if CompareRec(pRec1^.Data, pRec2^.Data, GroupL) = 0 then //key fields for the grouping are equal?
    begin
      Inc(avg);                         //calculation of the records for AVG and COUNT
      for j := 0 to AggL.Count - 1 do
      begin
        F := PVolgaQField(AggL.Items[j]);
        FI := FCursor.FindFieldInfo(F^.AsName);
        if FI <> nil then
          case F^.TokenType of
            vttCount: ;
            vttSum, vttAvg:             //adding two numbers (as double or integer) and put it on the place of the second
              SumBuf;
            vttMin:                     //copy the smaller value to the place of the greater (only for numbers and short strings!!!!!!!)
              if CompareBuf(pRec1^.Data + FI^.Offset, pRec2^.Data + FI^.Offset, FI, false) < 0 then
                Move((pRec1^.Data + FI^.Offset)^, (pRec2^.Data + FI^.Offset)^, FI^.DataSize);
            vttMax:                     //copy the greater value to the place of the smaller (only for numbers and short strings!!!!!!!)
              if CompareBuf(pRec1^.Data + FI^.Offset, pRec2^.Data + FI^.Offset, FI, false) > 0 then
                Move((pRec1^.Data + FI^.Offset)^, (pRec2^.Data + FI^.Offset)^, FI^.DataSize);
          end;
      end;
      //Delete the 1-st of the records
      pRec1^.Status := vusAbsent;
      FCursor.ClearData(pRec1^.Data);
    end
    else
    begin
      //if not comparison and AVG is given, then it is necessary to divide the summ on the count of the records in the group
      if HasAvg then
        SaveAvgCount;
      avg := 1;
    end;
  end;
  //for the last record
  if HasAvg and (FCursor.RecordCount >= 1) then
  begin
    pRec1 := FCursor.RecData[FCursor.RecordCount - 1];
    SaveAvgCount;
  end;
  GroupL.Free;
  AggL.Free;
end;

procedure TVolgaQuery.InternalApplyUpdates;
var T: PVolgaQTable;
  pRec: PVolgaRecord;
  i: integer;
begin
  //update master table
  T := PVolgaQTable(FTableList.Items[0]);
  if T^.IsMaster and T^.IsFile then
  begin
    FUniqueFI := TList.Create;
    GetFldInfoList(T^.QCursor, FUniqueFI, UniqueFieldNames);
    //we looking for the source in all records of the cursor using key fields in master-table and then edit the table
    for i := 0 to FCursor.RecordCount - 1 do
    begin
      pRec := FCursor.RecData[i];
      case pRec^.Status of
        vusModified: UpdateMaster(T, pRec);
        vusInserted: InsertMaster(T, pRec);
        vusDeleted: DeleteMaster(T, pRec);
      end;
    end;
    //save the original on disk when modification is over
    FUniqueFI.Free;
    FUniqueFI := nil;
    T^.QCursor.CommitUpdates;
    T^.QCursor.SaveToFile(T^.FullName, SaveOptions, '', false);
  end;
end;

procedure TVolgaQuery.InsertMaster(T: PVolgaQTable; pRec: PVolgaRecord);
//add new record into the master-table
var NewBuf: PChar;
  Msg: string;
begin
  NewBuf := AllocMem(T^.QCursor.RecSize);
  PrepareNewBuf(T, pRec, NewBuf);
  try
    if not T^.QCursor.SaveRecord(NewBuf, nil, Msg) then
      DatabaseError(Msg, Self);
  finally
    FreeMem(NewBuf);
  end;
end;

procedure TVolgaQuery.DeleteMaster(T: PVolgaQTable; pRec: PVolgaRecord);
var DelRec: PVolgaRecord;
  Msg: string;
begin
  DelRec := FindMaster(T, pRec);
  T^.QCursor.DeleteRecord(DelRec, Msg);
end;

procedure TVolgaQuery.UpdateMaster(T: PVolgaQTable; pRec: PVolgaRecord);
var UpdRec: PVolgaRecord;
  NewBuf: PChar;
  Msg: string;
  j: integer;
  FI: PVolgaField;
begin
  UpdRec := FindMaster(T, pRec);        //the record in master-table, that we shall change
  NewBuf := AllocMem(T^.QCursor.RecSize); //buffer for updated data storage
  //copy old values
  //copy the buffers of the long string-fields
  for j := 0 to T^.QCursor.FieldCount - 1 do
  begin
    FI := T^.QCursor.FieldInfo[j];
    case FI^.DataType of
      vftString:
        CopyLongString(UpdRec^.Data + FI^.Offset, NewBuf + FI^.Offset);
      vftMemo, vftFmtMemo, vftGraphic, vftBlob:
        CopyBlobField(UpdRec^.Data + FI^.Offset, NewBuf + FI^.Offset);
    else                                //the way of the copying of every field except string and BLOB
      Move((UpdRec^.Data + FI^.Offset)^, (NewBuf + FI^.Offset)^, FI^.DataSize);
    end;
  end;
  PrepareNewBuf(T, pRec, NewBuf);
  try
    if not T^.QCursor.SaveRecord(NewBuf, UpdRec, Msg) then
      DatabaseError(Msg, Self);
  finally
    FreeMem(NewBuf);
  end;
end;

function TVolgaQuery.FindMaster(T: PVolgaQTable; pRec: PVolgaRecord): PVolgaRecord;
var IsFound: Boolean;
  start, j: integer;
  FI, FM: PVolgaField;
begin
  Result := nil;
  start := 0;
  IsFound := false;
  while not IsFound and (start < T^.QCursor.RecordCount) do
  begin
    j := 0;
    IsFound := true;
    while IsFound and (j < FUniqueFI.Count) do
    begin
      FM := PVolgaField(FUniqueFI.Items[j]); //field from the master-table
      FI := FCursor.FindFieldInfo(FM^.Name);
      if pRec^.OldData <> nil then      //unmodified data
        IsFound := IsFound and (CompareBuf(pRec^.OldData + FI^.Offset,
          PVolgaRecord(T^.QCursor.RecData[start])^.Data + FM^.Offset, FM, false) = 0)
      else if PVolgaRecord(T^.QCursor.RecData[start])^.Status <> vusAbsent then
        IsFound := IsFound and (CompareBuf(pRec^.Data + FI^.Offset,
          PVolgaRecord(T^.QCursor.RecData[start])^.Data + FM^.Offset, FM, false) = 0);
      Inc(j);
    end;
    if not IsFound then
      Inc(start)
    else
    begin
      Result := PVolgaRecord(T^.QCursor.RecData[start]); //deleting of the found record
      Exit;
    end;
  end;
end;

procedure TVolgaQuery.PrepareNewBuf(T: PVolgaQTable; pRec: PVolgaRecord; NewBuf: PChar);
var iCol: integer;
  FI, FM: PVolgaField;
  sourbuf: PChar;
begin
  for iCol := 0 to T^.QCursor.FieldCount - 1 do
  begin
    FI := T^.QCursor.FieldInfo[iCol];   //field in the source table
    FM := FCursor.FindFieldInfo(FI^.Name); //field in the request
    if (FM <> nil) and (AnsiCompareText(FM^.Origin, Origin) = 0) then //field from our master-table?
    begin
      sourbuf := pRec^.Data + FM^.Offset; //changed buffer
      case FM^.DataType of
        vftString:
          CopyLongString(sourBuf, NewBuf + FI^.Offset);
        vftMemo, vftFmtMemo, vftGraphic, vftBlob:
          CopyBlobField(sourBuf, NewBuf + FI^.Offset);
      else
        Move(sourBuf^, (NewBuf + FI^.Offset)^, FI^.DataSize);
      end;
    end;
  end;
end;

procedure TVolgaQuery.SetOrigin(const Value: string);
begin
  inherited;
  //enter master-table token
//  if Active then begin
end;

procedure TVolgaQuery.AddDataset(Dataset: TVolgaDataset; const Alias: string);
var Item: TQueryDataset;
  s: string;
begin
  if not Assigned(Dataset) then
    Exit;
  if FQueryDatasets.IndexOfAlias(Alias) >= 0 then
    DatabaseError(V_DUPLICATEALIAS, Self);
  Item := FQueryDatasets.Add;
  s := Alias;
  if Length(s) = 0 then
    s := Dataset.Name;
  Item.Alias := s;
  Item.Dataset := Dataset;
end;

procedure TVolgaQuery.ClearDatasets;
begin
  FQueryDatasets.Clear;
end;

{ TQueryDataset }

procedure TQueryDataset.Assign(Source: TPersistent);
begin
  if Source is TQueryDataset then
  begin
    Alias := TQueryDataset(Source).Alias;
    Dataset := TQueryDataset(Source).Dataset;
  end
  else
    inherited Assign(Source);
end;

constructor TQueryDataset.Create(Collection: TCollection);
begin
  inherited Create(Collection);
end;

function TQueryDataset.GetDisplayName: string;
begin
  if FDataset <> nil then
    Result := FDataset.Name + ' (' + FAlias + ')'
  else
    Result := '(not defined)';
end;

procedure TQueryDataset.SetAlias(const Value: string);
begin
  if FAlias <> Value then
  begin
    if TQueryDatasets(Collection).IndexOfAlias(Value) >= 0 then
      DatabaseError(V_DUPLICATEALIAS, TQueryDatasets(Collection).FQuery);

    FAlias := Value;
    Changed(False);
  end;
end;

procedure TQueryDataset.SetDataset(Value: TVolgaDataset);
begin
  if Value = TQueryDatasets(Collection).FQuery then
    DatabaseError(V_CIRCULARREFERENCE, TQueryDatasets(Collection).FQuery);

  FDataset := Value;
  if (FAlias = '') and Assigned(Value) then
    FAlias := Value.Name;
end;

{ TQueryDatasets }

function TQueryDatasets.Add: TQueryDataset;
begin
  Result := TQueryDataset(inherited Add);
end;

constructor TQueryDatasets.Create(Query: TVolgaQuery);
begin
  inherited Create(TQueryDataset);
  FQuery := Query as TVolgaQuery;
end;

function TQueryDatasets.GetItem(Index: Integer): TQueryDataset;
begin
  Result := TQueryDataset(inherited GetItem(Index));
end;

function TQueryDatasets.GetOwner: TPersistent;
begin
  Result := FQuery;
end;

procedure TQueryDatasets.SetItem(Index: Integer; Value: TQueryDataset);
begin
  inherited SetItem(Index, Value);
end;

function TQueryDatasets.IndexOfAlias(const Alias: string): Integer;
var I: Integer;
begin
  Result := -1;
  for I := 0 to Count - 1 do
    if AnsiCompareText(TQueryDataset(inherited GetItem(I)).Alias, Alias) = 0 then
    begin
      Result := I;
      Exit;
    end;
end;

procedure TVolgaQuery.SetDatasets(const Value: TQueryDatasets);
begin
  FQueryDatasets.Assign(Value);
end;

end.

